/*
 * Copyright 2013 Square Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squareup.wire;

import com.squareup.javapoet.ClassName;
import com.squareup.javapoet.JavaFile;
import com.squareup.javapoet.TypeSpec;
import com.squareup.wire.java.JavaGenerator;
import com.squareup.wire.java.TypeWriter;
import com.squareup.wire.schema.Loader;
import com.squareup.wire.schema.Location;
import com.squareup.wire.schema.ProtoFile;
import com.squareup.wire.schema.Schema;
import com.squareup.wire.schema.Service;
import com.squareup.wire.schema.Type;
import java.io.IOException;

/** Compiler for Wire protocol buffers. */
public final class WireCompiler {
  private static final String CODE_GENERATED_BY_WIRE =
      "Code generated by Wire protocol buffer compiler, do not edit.";

  private final Loader loader;
  private final JavaGenerator.IO javaGeneratorIo;
  private final CommandLineOptions options;
  private final WireLogger log;

  /**
   * Runs the compiler.  See {@link CommandLineOptions} for command line options.
   */
  public static void main(String... args) {
    try {
      new WireCompiler(new CommandLineOptions(args)).compile();
    } catch (WireException e) {
      System.err.print("Fatal: ");
      e.printStackTrace(System.err);
      System.exit(1);
    }
  }

  WireCompiler(CommandLineOptions options) throws WireException {
    this(options, Loader.forSearchPaths(options.protoPaths()), JavaGenerator.IO.DEFAULT,
        new ConsoleWireLogger(options.quiet));
  }

  WireCompiler(CommandLineOptions options, Loader loader, JavaGenerator.IO javaGeneratorIo,
      WireLogger logger) throws WireException {
    this.options = options;
    this.loader = loader;
    this.javaGeneratorIo = javaGeneratorIo;
    this.log = logger;

    if (options.javaOut == null) {
      throw new WireException("Must specify " + CommandLineOptions.JAVA_OUT_FLAG + " flag");
    }
  }

  public void compile() throws WireException {
    Schema schema;
    try {
      schema = loader.load(options.sourceFileNames);
    } catch (IOException e) {
      throw new WireException(e);
    }

    if (!options.roots.isEmpty()) {
      log.info("Analyzing dependencies of root types.");
      schema = schema.retainRoots(options.roots);
    }

    JavaGenerator javaGenerator = JavaGenerator.get(schema);
    TypeWriter typeWriter = new TypeWriter(
        javaGenerator, options.emitOptions, options.enumOptions);

    for (ProtoFile protoFile : schema.protoFiles()) {
      if (!options.sourceFileNames.contains(protoFile.location().path())) {
        continue; // Don't emit anything for files not explicitly compiled.
      }

      for (Type type : protoFile.types()) {
        ClassName javaTypeName = (ClassName) javaGenerator.typeName(type.name());
        TypeSpec typeSpec = typeWriter.toTypeSpec(type);
        writeJavaFile(javaTypeName, typeSpec, type.location());
      }

      if (options.serviceFactory != null) {
        for (Service service : protoFile.services()) {
          TypeSpec typeSpec = options.serviceFactory.create(
              javaGenerator, options.serviceFactoryOptions, service);
          ClassName baseJavaTypeName = (ClassName) javaGenerator.typeName(service.name());
          // Use 'peerClass' to track service factories that add a prefix or suffix.
          ClassName generatedJavaTypeName = baseJavaTypeName.peerClass(typeSpec.name);
          writeJavaFile(generatedJavaTypeName, typeSpec, service.location());
        }
      }

      if (!protoFile.extendList().isEmpty()) {
        ClassName javaTypeName = javaGenerator.extensionsClass(protoFile);
        TypeSpec typeSpec = typeWriter.extensionsType(javaTypeName, protoFile);
        writeJavaFile(javaTypeName, typeSpec, protoFile.location());
      }
    }

    if (options.registryClass != null) {
      ClassName className = ClassName.bestGuess(options.registryClass);
      TypeSpec typeSpec = typeWriter.registryType(className, schema);
      writeJavaFile(className, typeSpec, null);
    }
  }

  private void writeJavaFile(
      ClassName javaTypeName, TypeSpec typeSpec, Location location) throws WireException {
    JavaFile.Builder builder = JavaFile.builder(javaTypeName.packageName(), typeSpec)
        .addFileComment("$L", CODE_GENERATED_BY_WIRE);
    if (location != null) {
      builder.addFileComment("\nSource file: $L", location);
    }
    JavaFile javaFile = builder.build();

    log.artifact(options.javaOut, javaFile);

    try {
      if (!options.dryRun) {
        javaGeneratorIo.write(options.javaOut, javaFile);
      }
    } catch (IOException e) {
      throw new WireException("Error emitting " + javaFile.packageName + "."
          + javaFile.typeSpec.name + " to " + options.javaOut, e);
    }
  }
}
